# -*- coding: utf-8 -*-

__author__ = "Marie Roald & Yngve Mardal Moe"
__version__ = "0.1.3"


from . import (
    data,
    factor_tools,
    model_evaluation,
    multimodel_evaluation,
    outliers,
    postprocessing,
    utils,
    visualisation,
)
